# Author: Simon Blanke
# Email: simon.blanke@yahoo.com
# License: MIT License

from typing import List, Dict, Literal, Union

from ..search import Search
from ..optimizers import ParticleSwarmOptimizer as _ParticleSwarmOptimizer


class ParticleSwarmOptimizer(_ParticleSwarmOptimizer, Search):
    """
    A class implementing **particle swarm optimization** for the public API.
    Inheriting from the `Search`-class to get the `search`-method and from
    the `ParticleSwarmOptimizer`-backend to get the underlying algorithm.

    Parameters
    ----------
    search_space : dict[str, list]
        The search space to explore. A dictionary with parameter
        names as keys and a numpy array as values.
    initialize : dict[str, int]
        The method to generate initial positions. A dictionary with
        the following key literals and the corresponding value type:
        {"grid": int, "vertices": int, "random": int, "warm_start": list[dict]}
    constraints : list[callable]
        A list of constraints, where each constraint is a callable.
        The callable returns `True` or `False` dependend on the input parameters.
    random_state : None, int
        If None, create a new random state. If int, create a new random state
        seeded with the value.
    rand_rest_p : float
        The probability of a random iteration during the the search process.
    population : int
        The number of particles in the swarm.
    inertia : float
        The inertia of the swarm.
    cognitive_weight : float
        A factor of the movement towards the personal best position of the individual optimizers in the population.
    social_weight : float
        A factor of the movement towards the personal best position of the individual optimizers in the population.
    temp_weight : float
        The temperature weight of the swarm.
    """

    def __init__(
        self,
        search_space: Dict[str, list],
        initialize: Dict[
            Literal["grid", "vertices", "random", "warm_start"],
            Union[int, list[dict]],
        ] = {"grid": 4, "random": 2, "vertices": 4},
        constraints: List[callable] = [],
        random_state: int = None,
        rand_rest_p: float = 0,
        nth_process: int = None,
        population: int = 10,
        inertia: float = 0.5,
        cognitive_weight: float = 0.5,
        social_weight: float = 0.5,
        temp_weight: float = 0.2,
    ):
        super().__init__(
            search_space=search_space,
            initialize=initialize,
            constraints=constraints,
            random_state=random_state,
            rand_rest_p=rand_rest_p,
            nth_process=nth_process,
            population=population,
            inertia=inertia,
            cognitive_weight=cognitive_weight,
            social_weight=social_weight,
            temp_weight=temp_weight,
        )
